@@ -5,12 +5,11 @@ local win32 = {}
 
 local fs = require("luarocks.fs")
 
-local cfg = require("luarocks.cfg")
+local cfg = require("luarocks.core.cfg")
 local dir = require("luarocks.dir")
+local path = require("luarocks.path")
 local util = require("luarocks.util")
 
-math.randomseed(os.time())
-
 -- Monkey patch io.popen and os.execute to make sure quoting
 -- works as expected.
 -- See http://lua-users.org/lists/lua-l/2013-11/msg00367.html
@@ -127,32 +126,54 @@ function win32.root_of(pathname)
 end
 
 --- Create a wrapper to make a script executable from the command-line.
--- @param file string: Pathname of script to be made executable.
--- @param dest string: Directory where to put the wrapper.
+-- @param script string: Pathname of script to be made executable.
+-- @param target string: wrapper target pathname (without wrapper suffix).
 -- @param name string: rock name to be used in loader context.
 -- @param version string: rock version to be used in loader context.
 -- @return boolean or (nil, string): True if succeeded, or nil and
 -- an error message.
-function win32.wrap_script(file, dest, name, version)
-   assert(type(file) == "string")
-   assert(type(dest) == "string")
+function win32.wrap_script(script, target, deps_mode, name, version, ...)
+   assert(type(script) == "string" or not script)
+   assert(type(target) == "string")
+   assert(type(deps_mode) == "string")
+   assert(type(name) == "string" or not name)
+   assert(type(version) == "string" or not version)
 
-   local base = dir.base_name(file)
-   local wrapname = fs.is_dir(dest) and dest.."/"..base or dest
-   wrapname = wrapname..".bat"
-   local lpath, lcpath = cfg.package_paths()
-   lpath = util.remove_path_dupes(lpath, ";")
-   lcpath = util.remove_path_dupes(lcpath, ";")
-   local wrapper = io.open(wrapname, "w")
+   local batname = target .. ".bat"
+   local wrapper = io.open(batname, "wb")
    if not wrapper then
-      return nil, "Could not open "..wrapname.." for writing."
+      return nil, "Could not open "..batname.." for writing."
    end
-   wrapper:write("@echo off\n")
-   local lua = dir.path(cfg.variables["LUA_BINDIR"], cfg.lua_interpreter)
-   local ppaths = "package.path="..util.LQ(lpath..";").."..package.path; package.cpath="..util.LQ(lcpath..";").."..package.cpath"
-   local addctx = "local k,l,_=pcall(require,"..util.LQ("luarocks.loader")..") _=k and l.add_context("..util.LQ(name)..","..util.LQ(version)..")"
-   wrapper:write(fs.Qb(lua)..' -e '..fs.Qb(ppaths)..' -e '..fs.Qb(addctx)..' '..fs.Qb(file)..' %*\n')
-   wrapper:write("exit /b %ERRORLEVEL%\n")
+
+   local lpath, lcpath = path.package_paths(deps_mode)
+   local lpath_var, lcpath_var = util.lua_path_variables()
+
+   local addctx
+   if name and version then
+      addctx = "local k,l,_=pcall(require,'luarocks.loader') _=k " ..
+                     "and l.add_context('"..name.."','"..version.."')"
+   end
+
+   local argv = {
+      fs.Qb(dir.path(cfg.variables["LUA_BINDIR"], cfg.lua_interpreter)),
+      script and fs.Qb(script) or "",
+      ...
+   }
+
+   wrapper:write("@echo off\r\n")
+   wrapper:write("set "..fs.Qb("LUAROCKS_SYSCONFDIR="..cfg.sysconfdir) .. "\r\n")
+   if target == "luarocks" or target == "luarocks-admin" then
+      wrapper:write("set "..fs.Qb(lpath_var.."="..package.path) .. "\r\n")
+      wrapper:write("set "..fs.Qb(lcpath_var.."="..package.cpath) .. "\r\n")
+   else
+      wrapper:write("set "..fs.Qb(lpath_var.."="..lpath..";%"..lpath_var.."%") .. "\r\n")
+      wrapper:write("set "..fs.Qb(lcpath_var.."="..lcpath..";%"..lcpath_var.."%") .. "\r\n")
+   end
+   if addctx then
+      wrapper:write("set "..fs.Qb("LUA_INIT=" .. addctx) .. "\r\n")
+   end
+   wrapper:write(table.concat(argv, " ") .. " %*\r\n")
+   wrapper:write("exit /b %ERRORLEVEL%\r\n")
    wrapper:close()
    return true
 end
@@ -187,10 +208,6 @@ function win32.copy_binary(filename, dest)
    return true
 end
 
-function win32.chmod(filename, mode)
-   return true
-end
-
 function win32.attributes(filename, attrtype)
    if attrtype == "permissions" then
       return "" -- FIXME
@@ -248,14 +265,14 @@ function win32.is_writable(file)
 end
 
 --- Create a temporary directory.
--- @param name string: name pattern to use for avoiding conflicts
+-- @param name_pattern string: name pattern to use for avoiding conflicts
 -- when creating temporary directory.
 -- @return string or (nil, string): name of temporary directory or (nil, error message) on failure.
-function win32.make_temp_dir(name)
-   assert(type(name) == "string")
-   name = dir.normalize(name)
+function win32.make_temp_dir(name_pattern)
+   assert(type(name_pattern) == "string")
+   name_pattern = dir.normalize(name_pattern)
 
-   local temp_dir = os.getenv("TMP") .. "/luarocks_" .. name:gsub(dir.separator, "_") .. "-" .. tostring(math.floor(math.random() * 10000))
+   local temp_dir = os.getenv("TMP") .. "/luarocks_" .. name_pattern:gsub("/", "_") .. "-" .. tostring(math.floor(math.random() * 10000))
    local ok, err = fs.make_dir(temp_dir)
    if ok then
       return temp_dir
@@ -272,4 +289,8 @@ function win32.current_user()
    return os.getenv("USERNAME")
 end
 
+function win32.export_cmd(var, val)
+   return ("SET %s=%s"):format(var, val)
+end
+
 return win32
