@@ -6,17 +6,22 @@ local tools = {}
 
 local fs = require("luarocks.fs")
 local dir = require("luarocks.dir")
-local cfg = require("luarocks.cfg")
+local cfg = require("luarocks.core.cfg")
 
-local vars = cfg.variables
+local vars = setmetatable({}, { __index = function(_,k) return cfg.variables[k] end })
 
 --- Adds prefix to command to make it run from a directory.
 -- @param directory string: Path to a directory.
 -- @param cmd string: A command-line string.
+-- @param exit_on_error bool: Exits immediately if entering the directory failed.
 -- @return string: The command-line with prefix.
-function tools.command_at(directory, cmd)
+function tools.command_at(directory, cmd, exit_on_error)
    local drive = directory:match("^([A-Za-z]:)")
-   cmd = "cd " .. fs.Q(directory) .. " & " .. cmd
+   local op = " & "
+   if exit_on_error then
+      op = " && "
+   end
+   local cmd = "cd " .. fs.Q(directory) .. op .. cmd
    if drive then
       cmd = drive .. " & " .. cmd
    end
@@ -79,7 +84,10 @@ end
 -- plus an error message.
 function tools.copy_contents(src, dest)
    assert(src and dest)
-   if fs.execute_quiet(fs.Q(vars.CP), "-dR", src.."\\*.*", dest) then
+   if not fs.is_dir(src) then
+      return false, src .. " is not a directory"
+   end
+   if fs.make_dir(dest) and fs.execute_quiet(fs.Q(vars.CP), "-dR", src.."\\*.*", dest) then
       return true
    else
       return false, "Failed copying "..src.." to "..dest
@@ -110,7 +118,7 @@ function tools.find(at)
       return {}
    end
    local result = {}
-   local pipe = io.popen(fs.command_at(at, fs.quiet_stderr(fs.Q(vars.FIND))))
+   local pipe = io.popen(fs.command_at(at, fs.quiet_stderr(fs.Q(vars.FIND)), true))
    for file in pipe:lines() do
       -- Windows find is a bit different
       local first_two = file:sub(1,2)
@@ -127,17 +135,70 @@ end
 -- @param zipfile string: pathname of .zip archive to be created.
 -- @param ... Filenames to be stored in the archive are given as
 -- additional arguments.
--- @return boolean: true on success, false on failure.
+-- @return boolean: true on success, nil and error message on failure.
 function tools.zip(zipfile, ...)
-   return fs.execute_quiet(fs.Q(vars.SEVENZ).." -aoa a -tzip", zipfile, ...)
+   if fs.execute_quiet(fs.Q(vars.SEVENZ).." -aoa a -tzip", zipfile, ...) then
+      return true
+   else
+      return nil, "failed compressing " .. zipfile
+   end
 end
 
 --- Uncompress files from a .zip archive.
 -- @param zipfile string: pathname of .zip archive to be extracted.
--- @return boolean: true on success, false on failure.
+-- @return boolean: true on success, nil and error message on failure.
 function tools.unzip(zipfile)
    assert(zipfile)
-   return fs.execute_quiet(fs.Q(vars.SEVENZ).." -aoa x", zipfile)
+   if fs.execute_quiet(fs.Q(vars.SEVENZ).." -aoa x", zipfile) then
+      return true
+   else
+      return nil, "failed extracting " .. zipfile
+   end
+end
+
+local function sevenz(default_ext, infile, outfile)
+   assert(type(infile) == "string")
+   assert(outfile == nil or type(outfile) == "string")
+
+   local dropext = infile:gsub("%."..default_ext.."$", "")
+   local outdir = dir.dir_name(dropext)
+
+   infile = fs.absolute_name(infile)
+
+   local cmdline = fs.Q(vars.SEVENZ).." -aoa -t* -o"..fs.Q(outdir).." x "..fs.Q(infile)
+   local ok, err = fs.execute_quiet(cmdline)
+   if not ok then
+      return nil, "failed extracting " .. infile
+   end
+
+   if outfile then
+      outfile = fs.absolute_name(outfile)
+      dropext = fs.absolute_name(dropext)
+      ok, err = os.rename(dropext, outfile)
+      if not ok then
+         return nil, "failed creating new file " .. outfile
+      end
+   end
+
+   return true
+end
+
+--- Uncompresses a .gz file.
+-- @param infile string: pathname of .gz file to be extracted.
+-- @param outfile string or nil: pathname of output file to be produced.
+-- If not given, name is derived from input file.
+-- @return boolean: true on success; nil and error message on failure.
+function tools.gunzip(infile, outfile)
+   return sevenz("gz", infile, outfile)
+end
+
+--- Uncompresses a .bz2 file.
+-- @param infile string: pathname of .bz2 file to be extracted.
+-- @param outfile string or nil: pathname of output file to be produced.
+-- If not given, name is derived from input file.
+-- @return boolean: true on success; nil and error message on failure.
+function tools.bunzip2(infile, outfile)
+   return sevenz("bz2", infile, outfile)
 end
 
 --- Test is pathname is a directory.
@@ -156,59 +217,78 @@ function tools.is_file(file)
    return fs.execute(fs.Q(vars.TEST).." -f", file)
 end
 
---- Strip the last extension of a filename.
--- Example: "foo.tar.gz" becomes "foo.tar".
--- If filename has no dots, returns it unchanged.
--- @param filename string: The file name to strip.
--- @return string: The stripped name.
-local function strip_extension(filename)
-   assert(type(filename) == "string")
-   return (filename:gsub("%.[^.]+$", "")) or filename
+--- Helper function for fs.set_permissions
+-- @return table: an array of all system users
+local function get_system_users()
+   local result = {}
+   local fd = assert(io.popen("wmic UserAccount get name"))
+   for user in fd:lines() do
+      user = user:gsub("%s+$", "")
+      if user ~= "" and user ~= "Name" and user ~= "Administrator" then
+         table.insert(result, user)
+      end
+   end
+   return result
 end
 
---- Uncompress gzip file.
--- @param archive string: Filename of archive.
--- @return boolean : success status
-local function gunzip(archive)
-  return fs.execute_quiet(fs.Q(vars.SEVENZ).." -aoa x", archive)
-end
+--- Set permissions for file or directory
+-- @param filename string: filename whose permissions are to be modified
+-- @param mode string ("read" or "exec"): permission to set
+-- @param scope string ("user" or "all"): the user(s) to whom the permission applies
+-- @return boolean or (boolean, string): true on success, false on failure,
+-- plus an error message
+function tools.set_permissions(filename, mode, scope)
+   assert(filename and mode and scope)
 
---- Unpack an archive.
--- Extract the contents of an archive, detecting its format by
--- filename extension.
--- @param archive string: Filename of archive.
--- @return boolean or (boolean, string): true on success, false and an error message on failure.
-function tools.unpack_archive(archive)
-   assert(type(archive) == "string")
+   if scope == "user" then
+      local perms
+      if mode == "read" then
+         perms = "(R,W,M)"
+      elseif mode == "exec" then
+         perms = "(F)"
+      end
 
-   local ok
-   local sevenzx = fs.Q(vars.SEVENZ).." -aoa x"
-   if archive:match("%.tar%.gz$") then
-      ok = gunzip(archive)
-      if ok then
-         ok = fs.execute_quiet(sevenzx, strip_extension(archive))
+      local ok
+      -- Take ownership of the given file
+      ok = fs.execute_quiet("takeown /f " .. fs.Q(filename))
+      if not ok then
+         return false, "Could not take ownership of the given file"
       end
-   elseif archive:match("%.tgz$") then
-      ok = gunzip(archive)
-      if ok then
-         ok = fs.execute_quiet(sevenzx, strip_extension(archive)..".tar")
+      -- Grant the current user the proper rights
+      ok = fs.execute_quiet(fs.Q(vars.ICACLS) .. " " .. fs.Q(filename) .. " /inheritance:d /grant:r %USERNAME%:" .. perms)
+      if not ok then
+         return false, "Failed setting permission " .. mode .. " for " .. scope
       end
-   elseif archive:match("%.tar%.bz2$") then
-      ok = fs.execute_quiet(sevenzx, archive)
-      if ok then
-         ok = fs.execute_quiet(sevenzx, strip_extension(archive))
+      -- Finally, remove all the other users from the ACL in order to deny them access to the file
+      for _, user in pairs(get_system_users()) do
+         local ok = fs.execute_quiet(fs.Q(vars.ICACLS) .. " " .. fs.Q(filename) .. " /remove " .. fs.Q(user))
+         if not ok then
+            return false, "Failed setting permission " .. mode .. " for " .. scope
+         end
+      end
+   elseif scope == "all" then
+      local my_perms, others_perms
+      if mode == "read" then
+         my_perms = "(R,W,M)"
+         others_perms = "(R)"
+      elseif mode == "exec" then
+         my_perms = "(F)"
+         others_perms = "(RX)"
+      end
+
+      local ok
+      -- Grant permissions available to all users
+      ok = fs.execute_quiet(fs.Q(vars.ICACLS) .. " " .. fs.Q(filename) .. " /inheritance:d /grant:r Everyone:" .. others_perms)
+      if not ok then
+         return false, "Failed setting permission " .. mode .. " for " .. scope
+      end
+      -- Grant permissions available only to the current user
+      ok = fs.execute_quiet(fs.Q(vars.ICACLS) .. " " .. fs.Q(filename) .. " /inheritance:d /grant %USERNAME%:" .. my_perms)
+      if not ok then
+         return false, "Failed setting permission " .. mode .. " for " .. scope
       end
-   elseif archive:match("%.zip$") then
-      ok = fs.execute_quiet(sevenzx, archive)
-   elseif archive:match("%.lua$") or archive:match("%.c$") then
-      -- Ignore .lua and .c files; they don't need to be extracted.
-      return true
-   else
-      return false, "Couldn't extract archive "..archive..": unrecognized filename extension"
-   end
-   if not ok then
-      return false, "Failed extracting "..archive
    end
+
    return true
 end
 
@@ -224,4 +304,13 @@ function tools.browser(url)
    return fs.execute(cfg.web_browser..' "Starting docs..." '..fs.Q(url))
 end
 
+-- Set access and modification times for a file.
+-- @param filename File to set access and modification times for.
+-- @param time may be a string or number containing the format returned
+-- by os.time, or a table ready to be processed via os.time; if
+-- nil, current time is assumed.
+function tools.set_time(filename, time)
+   return true -- FIXME
+end
+
 return tools
