@@ -1,31 +1,20 @@
 
---- Module implementing the LuaRocks "pack" command.
--- Creates a rock, packing sources or binaries.
---module("luarocks.pack", package.seeall)
+-- Create rock files, packing sources or binaries.
 local pack = {}
-package.loaded["luarocks.pack"] = pack
 
 local unpack = unpack or table.unpack
 
+local queries = require("luarocks.queries")
 local path = require("luarocks.path")
 local repos = require("luarocks.repos")
 local fetch = require("luarocks.fetch")
 local fs = require("luarocks.fs")
-local cfg = require("luarocks.cfg")
+local cfg = require("luarocks.core.cfg")
 local util = require("luarocks.util")
 local dir = require("luarocks.dir")
 local manif = require("luarocks.manif")
 local search = require("luarocks.search")
 
-pack.help_summary = "Create a rock, packing sources or binaries."
-pack.help_arguments = "{<rockspec>|<name> [<version>]}"
-pack.help = [[
-Argument may be a rockspec file, for creating a source rock,
-or the name of an installed package, for creating a binary rock.
-In the latter case, the app version may be given as a second
-argument.
-]]
-
 --- Create a source rock.
 -- Packages a rockspec and its required source files in a rock
 -- file with the .src.rock extension, which can later be built and
@@ -40,7 +29,7 @@ function pack.pack_source_rock(rockspec_file)
    if err then
       return nil, "Error loading rockspec: "..err
    end
-   rockspec_file = rockspec.local_filename
+   rockspec_file = rockspec.local_abs_filename
 
    local name_version = rockspec.name .. "-" .. rockspec.version
    local rock_file = fs.absolute_name(name_version .. ".src.rock")
@@ -53,16 +42,17 @@ function pack.pack_source_rock(rockspec_file)
    if not ok then return nil, err end
 
    fs.delete(rock_file)
-   fs.copy(rockspec_file, source_dir)
-   if not fs.zip(rock_file, dir.base_name(rockspec_file), dir.base_name(source_file)) then
-      return nil, "Failed packing "..rock_file
+   fs.copy(rockspec_file, source_dir, "read")
+   ok, err = fs.zip(rock_file, dir.base_name(rockspec_file), dir.base_name(source_file))
+   if not ok then
+      return nil, "Failed packing "..rock_file.." - "..err
    end
    fs.pop_dir()
 
    return rock_file
 end
 
-local function copy_back_files(name, version, file_tree, deploy_dir, pack_dir)
+local function copy_back_files(name, version, file_tree, deploy_dir, pack_dir, perms)
    local ok, err = fs.make_dir(pack_dir)
    if not ok then return nil, err end
    for file, sub in pairs(file_tree) do
@@ -74,9 +64,9 @@ local function copy_back_files(name, version, file_tree, deploy_dir, pack_dir)
       else
          local versioned = path.versioned_name(source, deploy_dir, name, version)
          if fs.exists(versioned) then
-            fs.copy(versioned, target)
+            fs.copy(versioned, target, perms)
          else
-            fs.copy(source, target)
+            fs.copy(source, target, perms)
          end
       end
    end
@@ -85,47 +75,24 @@ end
 
 -- @param name string: Name of package to pack.
 -- @param version string or nil: A version number may also be passed.
+-- @param tree string or nil: An optional tree to pick the package from.
 -- @return string or (nil, string): The filename of the resulting
 -- .src.rock file; or nil and an error message.
-local function do_pack_binary_rock(name, version)
-   assert(type(name) == "string")
-   assert(type(version) == "string" or not version)
+function pack.pack_installed_rock(query, tree)
 
-   local query = search.make_query(name, version)
-   query.exact_name = true
-   local results = {}
-   
-   search.manifest_search(results, cfg.rocks_dir, query)
-   
-   if not next(results) then
-      return nil, "'"..name.."' does not seem to be an installed rock."
+   local name, version, repo, repo_url = search.pick_installed_rock(query, tree)
+   if not name then
+      return nil, version
    end
-   
-   local versions = results[name]
-   
-   if not version then
-      local first = next(versions)
-      if next(versions, first) then
-         return nil, "Please specify which version of '"..name.."' to pack."
-      end
-      version = first
-   end
-   if not version:match("[^-]+%-%d+") then
-      return nil, "Expected version "..version.." in version-revision format."
-   end
-   
-   local info = versions[version][1]
-   
-   local root = path.root_dir(info.repo)
+
+   local root = path.root_from_rocks_dir(repo_url)
    local prefix = path.install_dir(name, version, root)
    if not fs.exists(prefix) then
       return nil, "'"..name.." "..version.."' does not seem to be an installed rock."
    end
    
-   local rock_manifest = manif.load_rock_manifest(name, version, root)
-   if not rock_manifest then
-      return nil, "rock_manifest file not found for "..name.." "..version.." - not a LuaRocks 2 tree?"
-   end
+   local rock_manifest, err = manif.load_rock_manifest(name, version, root)
+   if not rock_manifest then return nil, err end
 
    local name_version = name .. "-" .. version
    local rock_file = fs.absolute_name(name_version .. "."..cfg.arch..".rock")
@@ -135,12 +102,12 @@ local function do_pack_binary_rock(name, version)
 
    local is_binary = false
    if rock_manifest.lib then
-      local ok, err = copy_back_files(name, version, rock_manifest.lib, path.deploy_lib_dir(root), dir.path(temp_dir, "lib"))
+      local ok, err = copy_back_files(name, version, rock_manifest.lib, path.deploy_lib_dir(repo), dir.path(temp_dir, "lib"), "exec")
       if not ok then return nil, "Failed copying back files: " .. err end
       is_binary = true
    end
    if rock_manifest.lua then
-      local ok, err = copy_back_files(name, version, rock_manifest.lua, path.deploy_lua_dir(root), dir.path(temp_dir, "lua"))
+      local ok, err = copy_back_files(name, version, rock_manifest.lua, path.deploy_lua_dir(repo), dir.path(temp_dir, "lua"), "read")
       if not ok then return nil, "Failed copying back files: " .. err end
    end
    
@@ -158,7 +125,7 @@ local function do_pack_binary_rock(name, version)
    return rock_file
 end
 
-function pack.pack_binary_rock(name, version, cmd, ...)
+function pack.pack_binary_rock(name, version, cmd)
 
    -- The --pack-binary-rock option for "luarocks build" basically performs
    -- "luarocks build" on a temporary tree and then "luarocks pack". The
@@ -174,7 +141,7 @@ function pack.pack_binary_rock(name, version, cmd, ...)
    util.schedule_function(fs.delete, temp_dir)
 
    path.use_tree(temp_dir)
-   local ok, err = cmd(...)
+   local ok, err = cmd()
    if not ok then
       return nil, err
    end
@@ -182,35 +149,8 @@ function pack.pack_binary_rock(name, version, cmd, ...)
    if not rname then
       rname, rversion = name, version
    end
-   return do_pack_binary_rock(rname, rversion)
-end
-
---- Driver function for the "pack" command.
--- @param arg string:  may be a rockspec file, for creating a source rock,
--- or the name of an installed package, for creating a binary rock.
--- @param version string or nil: if the name of a package is given, a
--- version may also be passed.
--- @return boolean or (nil, string): true if successful or nil followed
--- by an error message.
-function pack.run(...)
-   local flags, arg, version = util.parse_flags(...)
-   assert(type(version) == "string" or not version)
-   if type(arg) ~= "string" then
-      return nil, "Argument missing. "..util.see_help("pack")
-   end
-
-   local file, err
-   if arg:match(".*%.rockspec") then
-      file, err = pack.pack_source_rock(arg)
-   else
-      file, err = do_pack_binary_rock(arg, version)
-   end
-   if err then
-      return nil, err
-   else
-      util.printout("Packed: "..file)
-      return true
-   end
+   local query = queries.new(rname, rversion)
+   return pack.pack_installed_rock(query, temp_dir)
 end
 
 return pack
