@@ -4,82 +4,11 @@
 -- Implemented separately to avoid interdependencies,
 -- as it is used in the bootstrapping stage of the cfg module.
 local persist = {}
-package.loaded["luarocks.persist"] = persist
 
+local core = require("luarocks.core.persist")
 local util = require("luarocks.util")
 
---- Load and run a Lua file in an environment.
--- @param filename string: the name of the file.
--- @param env table: the environment table.
--- @return (true, any) or (nil, string, string): true and the return value
--- of the file, or nil, an error message and an error code ("open", "load"
--- or "run") in case of errors.
-local function run_file(filename, env)
-   local fd, err = io.open(filename)
-   if not fd then
-      return nil, err, "open"
-   end
-   local str, err = fd:read("*a")
-   fd:close()
-   if not str then
-      return nil, err, "open"
-   end
-   str = str:gsub("^#![^\n]*\n", "")
-   local chunk, ran
-   if _VERSION == "Lua 5.1" then -- Lua 5.1
-      chunk, err = loadstring(str, filename)
-      if chunk then
-         setfenv(chunk, env)
-         ran, err = pcall(chunk)
-      end
-   else -- Lua 5.2
-      chunk, err = load(str, filename, "t", env)
-      if chunk then
-         ran, err = pcall(chunk)
-      end
-   end
-   if not chunk then
-      return nil, "Error loading file: "..err, "load"
-   end
-   if not ran then
-      return nil, "Error running file: "..err, "run"
-   end
-   return true, err
-end
-
---- Load a Lua file containing assignments, storing them in a table.
--- The global environment is not propagated to the loaded file.
--- @param filename string: the name of the file.
--- @param tbl table or nil: if given, this table is used to store
--- loaded values.
--- @return (table, table) or (nil, string, string): a table with the file's assignments
--- as fields and set of undefined globals accessed in file,
--- or nil, an error message and an error code ("open"; couldn't open the file,
--- "load"; compile-time error, or "run"; run-time error)
--- in case of errors.
-function persist.load_into_table(filename, tbl)
-   assert(type(filename) == "string")
-   assert(type(tbl) == "table" or not tbl)
-
-   local result = tbl or {}
-   local globals = {}
-   local globals_mt = {
-      __index = function(t, k)
-         globals[k] = true
-      end
-   }
-   local save_mt = getmetatable(result)
-   setmetatable(result, globals_mt)
-   
-   local ok, err, errcode = run_file(filename, result)
-   
-   setmetatable(result, save_mt)
-
-   if not ok then
-      return nil, err, errcode
-   end
-   return result, globals
-end
+persist.load_into_table = core.load_into_table
 
 local write_table
 
@@ -129,7 +58,7 @@ write_table = function(out, tbl, level, field_order)
    for k, v, sub_order in util.sortedpairs(tbl, field_order) do
       out:write(sep)
       if indent then
-         for n = 1,level do out:write(indentation) end
+         for _ = 1, level do out:write(indentation) end
       end
 
       if k == i then
@@ -157,7 +86,7 @@ write_table = function(out, tbl, level, field_order)
    end
    if sep ~= "\n" then
       out:write("\n")
-      for n = 1,level-1 do out:write(indentation) end
+      for _ = 1, level - 1 do out:write(indentation) end
    end
    out:write("}")
 end
@@ -174,6 +103,19 @@ local function write_table_as_assignments(out, tbl, field_order)
    end
 end
 
+--- Write a table as series of assignments to a writer object.
+-- @param out table or userdata: a writer object supporting :write() method.
+-- @param tbl table: the table to be written.
+local function write_table_as_table(out, tbl)
+   out:write("return {\n")
+   for k, v, sub_order in util.sortedpairs(tbl) do
+      out:write("   " .. k .. " = ")
+      write_value(out, v, 1, sub_order)
+      out:write(",\n")
+   end
+   out:write("}\n")
+end
+
 --- Save the contents of a table to a string.
 -- Each element of the table is saved as a global assignment.
 -- Only numbers, strings and tables (containing numbers, strings
@@ -207,4 +149,22 @@ function persist.save_from_table(filename, tbl, field_order)
    return true
 end
 
+--- Save the contents of a table as a module.
+-- Each element of the table is saved as a global assignment.
+-- Only numbers, strings and tables (containing numbers, strings
+-- or other recursively processed tables) are supported.
+-- @param filename string: the output filename
+-- @param tbl table: the table containing the data to be written
+-- @return boolean or (nil, string): true if successful, or nil and a
+-- message in case of errors.
+function persist.save_as_module(filename, tbl)
+   local out = io.open(filename, "w")
+   if not out then
+      return nil, "Cannot create file at "..filename
+   end
+   write_table_as_table(out, tbl)
+   out:close()
+   return true
+end
+
 return persist
