@@ -1,8 +1,9 @@
 
 local api = {}
 
-local cfg = require("luarocks.cfg")
+local cfg = require("luarocks.core.cfg")
 local fs = require("luarocks.fs")
+local dir = require("luarocks.dir")
 local util = require("luarocks.util")
 local persist = require("luarocks.persist")
 local multipart = require("luarocks.upload.multipart")
@@ -20,8 +21,8 @@ end
 function Api:load_config()
    local upload_conf = upload_config_file()
    if not upload_conf then return nil end
-   local cfg, err = persist.load_into_table(upload_conf)
-   return cfg
+   local config, err = persist.load_into_table(upload_conf)
+   return config
 end
 
 function Api:save_config()
@@ -36,8 +37,12 @@ function Api:save_config()
    end
    local upload_conf = upload_config_file()
    if not upload_conf then return nil end
+   local ok, err = fs.make_dir(dir.dir_name(upload_conf))
+   if not ok then
+      return nil, err
+   end
    persist.save_from_table(upload_conf, self.config)
-   fs.chmod(upload_conf, "0600")
+   fs.set_permissions(upload_conf, "read", "user")
 end
 
 function Api:check_version()
@@ -57,7 +62,7 @@ function Api:check_version()
          return nil, "Your upload client is too out of date to continue, please upgrade LuaRocks."
       end
       if res.version ~= tool_version then
-         util.printerr("Warning: Your LuaRocks is out of date, consider upgrading.")
+         util.warning("your LuaRocks is out of date, consider upgrading.")
       end
    end
    return true
@@ -110,6 +115,7 @@ local function require_json()
    for _, lib in ipairs(list) do
       local json_ok, json = pcall(require, lib)
       if json_ok then
+         pcall(json.use_lpeg) -- optional feature in dkjson
          return json_ok, json
       end
    end
@@ -137,7 +143,7 @@ function Api:request(url, params, post_params)
    local json_ok, json = require_json()
    if not json_ok then return nil, "A JSON library is required for this command. "..json end
    
-   if cfg.downloader == "wget" then
+   if fs.which_tool("downloader") == "wget" then
       local curl_ok, err = fs.is_tool_available(vars.CURL, "curl")
       if not curl_ok then
          return nil, err
@@ -147,8 +153,6 @@ function Api:request(url, params, post_params)
    if not self.config.key then
       return nil, "Must have API key before performing any actions."
    end
-   local body
-   local headers = {}
    if params and next(params) then
       url = url .. ("?" .. encode_query_string(params))
    end
@@ -215,7 +219,6 @@ function Api:request(url, params, post_params)
             warned_luasec = true
          end
          http_ok, http = pcall(require, "socket.http")
-         server = server:gsub("^https", "http")
          url = url:gsub("^https", "http")
          via = "luasocket"
       end
@@ -269,7 +272,7 @@ function api.new(flags)
    self.config = self:load_config() or {}
    self.config.server = flags["server"] or self.config.server or cfg.upload.server
    self.config.version = self.config.version or cfg.upload.version
-   self.config.key = flags["api-key"] or self.config.key
+   self.config.key = flags["temp-key"] or flags["api-key"] or self.config.key
    self.debug = flags["debug"]
    if not self.config.key then
       return nil, "You need an API key to upload rocks.\n" ..
@@ -283,4 +286,3 @@ function api.new(flags)
 end
 
 return api
-
