@@ -2,10 +2,23 @@
 --- Assorted utilities for managing tables, plus a scheduler for rollback functions.
 -- Does not requires modules directly (only as locals
 -- inside specific functions) to avoid interdependencies,
--- as this is used in the bootstrapping stage of luarocks.cfg.
+-- as this is used in the bootstrapping stage of luarocks.core.cfg.
 
 local util = {}
 
+local core = require("luarocks.core.util")
+
+util.cleanup_path = core.cleanup_path
+util.split_string = core.split_string
+util.sortedpairs = core.sortedpairs
+util.deep_merge = core.deep_merge
+util.deep_merge_under = core.deep_merge_under
+util.popen_read = core.popen_read
+util.show_table = core.show_table
+util.printerr = core.printerr
+util.warning = core.warning
+util.keys = core.keys
+
 local unpack = unpack or table.unpack
 
 local scheduled_functions = {}
@@ -46,7 +59,9 @@ end
 -- Functions are executed in the inverse order they were scheduled.
 function util.run_scheduled_functions()
    local fs = require("luarocks.fs")
-   fs.change_dir_to_root()
+   if fs.change_dir_to_root then
+      fs.change_dir_to_root()
+   end
    for i = #scheduled_functions, 1, -1 do
       local item = scheduled_functions[i]
       item.fn(unpack(item.args))
@@ -76,17 +91,22 @@ local supported_flags = {
    ["bin"] = true,
    ["binary"] = true,
    ["branch"] = "<branch-name>",
+   ["build-deps"] = true,
    ["debug"] = true,
    ["deps"] = true,
    ["deps-mode"] = "<mode>",
    ["detailed"] = "\"<text>\"",
+   ["dev"] = true,
    ["force"] = true,
    ["force-fast"] = true,
    ["from"] = "<server>",
    ["help"] = true,
    ["home"] = true,
    ["homepage"] = "\"<url>\"",
+   ["index"] = true,
+   ["issues"] = true,
    ["keep"] = true,
+   ["labels"] = true,
    ["lib"] = "<library>",
    ["license"] = "\"<text>\"",
    ["list"] = true,
@@ -95,12 +115,16 @@ local supported_flags = {
    ["lr-bin"] = true,
    ["lr-cpath"] = true,
    ["lr-path"] = true,
+   ["lua-dir"] = "<path>",
    ["lua-version"] = "<vers>",
+   ["lua-versions"] = "<versions>",
    ["lua-ver"] = true,
    ["lua-incdir"] = true,
    ["lua-libdir"] = true,
    ["modules"] = true,
    ["mversion"] = true,
+   ["namespace"] = "<namespace>",
+   ["no-bin"] = true,
    ["no-refresh"] = true,
    ["nodeps"] = true,
    ["old-versions"] = true,
@@ -113,8 +137,12 @@ local supported_flags = {
    ["output"] = "<file>",
    ["pack-binary-rock"] = true,
    ["porcelain"] = true,
+   ["project-tree"] = "<tree>",
    ["quick"] = true,
+   ["reset"] = true,
    ["rock-dir"] = true,
+   ["rock-license"] = true,
+   ["rock-namespace"] = true,
    ["rock-tree"] = true,
    ["rock-trees"] = true,
    ["rockspec"] = true,
@@ -125,6 +153,8 @@ local supported_flags = {
    ["summary"] = "\"<text>\"",
    ["system-config"] = true,
    ["tag"] = "<tag>",
+   ["test-type"] = "<type>",
+   ["temp-key"] = "<key>",
    ["timeout"] = "<seconds>",
    ["to"] = "<path>",
    ["tree"] = "<path>",
@@ -143,13 +173,12 @@ function util.parse_flags(...)
    local flags = {}
    local i = 1
    local out = {}
-   local ignore_flags = false
+   local state = "initial"
    while i <= #args do
       local flag = args[i]:match("^%-%-(.*)")
-      if flag == "--" then
-         ignore_flags = true
-      end
-      if flag and not ignore_flags then
+      if state == "initial" and flag == "" then
+         state = "ignore_flags"
+      elseif state == "initial" and flag then
          local var,val = flag:match("([a-z_%-]*)=(.*)")
          if val then
             local vartype = supported_flags[var]
@@ -188,7 +217,7 @@ function util.parse_flags(...)
                return { ERROR = "Invalid argument: unknown flag --"..var.."." }
             end
          end
-      else
+      elseif state == "ignore_flags" or (state == "initial" and not flag) then
          table.insert(out, args[i])
       end
       i = i + 1
@@ -196,80 +225,8 @@ function util.parse_flags(...)
    return flags, unpack(out)
 end
 
--- Adds legacy 'run' function to a command module.
--- @param command table: command module with 'command' function,
--- the added 'run' function calls it after parseing command-line arguments.
-function util.add_run_function(command)
-   command.run = function(...) return command.command(util.parse_flags(...)) end
-end
-
---- Merges contents of src on top of dst's contents.
--- @param dst Destination table, which will receive src's contents.
--- @param src Table which provides new contents to dst.
--- @see platform_overrides
-function util.deep_merge(dst, src)
-   for k, v in pairs(src) do
-      if type(v) == "table" then
-         if not dst[k] then
-            dst[k] = {}
-         end
-         if type(dst[k]) == "table" then
-            util.deep_merge(dst[k], v)
-         else
-            dst[k] = v
-         end
-      else
-         dst[k] = v
-      end
-   end
-end
-
---- Perform platform-specific overrides on a table.
--- Overrides values of table with the contents of the appropriate
--- subset of its "platforms" field. The "platforms" field should
--- be a table containing subtables keyed with strings representing
--- platform names. Names that match the contents of the global
--- cfg.platforms setting are used. For example, if
--- cfg.platforms= {"foo"}, then the fields of
--- tbl.platforms.foo will overwrite those of tbl with the same
--- names. For table values, the operation is performed recursively
--- (tbl.platforms.foo.x.y.z overrides tbl.x.y.z; other contents of
--- tbl.x are preserved).
--- @param tbl table or nil: Table which may contain a "platforms" field;
--- if it doesn't (or if nil is passed), this function does nothing.
-function util.platform_overrides(tbl)
-   assert(type(tbl) == "table" or not tbl)
-   
-   local cfg = require("luarocks.cfg")
-   
-   if not tbl then return end
-   
-   if tbl.platforms then
-      for _, platform in ipairs(cfg.platforms) do
-         local platform_tbl = tbl.platforms[platform]
-         if platform_tbl then
-            util.deep_merge(tbl, platform_tbl)
-         end
-      end
-   end
-   tbl.platforms = nil
-end
-
 local var_format_pattern = "%$%((%a[%a%d_]+)%)"
 
---- Create a new shallow copy of a table: a new table with
--- the same keys and values. Keys point to the same objects as
--- the original table (ie, does not copy recursively).
--- @param tbl table: the input table
--- @return table: a new table with the same contents.
-function util.make_shallow_copy(tbl)
-   local copy = {}
-   for k,v in pairs(tbl) do
-      copy[k] = v
-   end
-   return copy
-end
-
 -- Check if a set of needed variables are referenced
 -- somewhere in a list of definitions, warning the user
 -- about any unused ones. Each key in needed_set should
@@ -281,7 +238,7 @@ end
 -- needed variables.
 -- @param msg string: the warning message to display.
 function util.warn_if_not_used(var_defs, needed_set, msg)
-   needed_set = util.make_shallow_copy(needed_set)
+   needed_set = core.make_shallow_copy(needed_set)
    for _, val in pairs(var_defs) do
       for used in val:gmatch(var_format_pattern) do
          needed_set[used] = nil
@@ -332,95 +289,40 @@ function util.variable_substitutions(tbl, vars)
    end
 end
 
---- Return an array of keys of a table.
--- @param tbl table: The input table.
--- @return table: The array of keys.
-function util.keys(tbl)
-   local ks = {}
-   for k,_ in pairs(tbl) do
-      table.insert(ks, k)
-   end
-   return ks
-end
-
-local function default_sort(a, b)
-   local ta = type(a)
-   local tb = type(b)
-   if ta == "number" and tb == "number" then
-      return a < b
-   elseif ta == "number" then
-      return true
-   elseif tb == "number" then
-      return false
-   else
-      return tostring(a) < tostring(b)
-   end
-end
-
---- A table iterator generator that returns elements sorted by key,
--- to be used in "for" loops.
--- @param tbl table: The table to be iterated.
--- @param sort_function function or table or nil: An optional comparison function
--- to be used by table.sort when sorting keys, or an array listing an explicit order
--- for keys. If a value itself is an array, it is taken so that the first element
--- is a string representing the field name, and the second element is a priority table
--- for that key, which is returned by the iterator as the third value after the key
--- and the value.
--- @return function: the iterator function.
-function util.sortedpairs(tbl, sort_function)
-   sort_function = sort_function or default_sort
-   local keys = util.keys(tbl)
-   local sub_orders = {}
-
-   if type(sort_function) == "function" then
-      table.sort(keys, sort_function)
-   else
-      local order = sort_function
-      local ordered_keys = {}
-      local all_keys = keys
-      keys = {}
-
-      for _, order_entry in ipairs(order) do
-         local key, sub_order
-         if type(order_entry) == "table" then
-            key = order_entry[1]
-            sub_order = order_entry[2]
-         else
-            key = order_entry
-         end
-
-         if tbl[key] then
-            ordered_keys[key] = true
-            sub_orders[key] = sub_order
-            table.insert(keys, key)
-         end
-      end
-
-      table.sort(all_keys, default_sort)
-      for _, key in ipairs(all_keys) do
-         if not ordered_keys[key] then
-            table.insert(keys, key)
-         end
-      end
-   end
-
-   local i = 1
-   return function()
-      local key = keys[i]
-      i = i + 1
-      return key, tbl[key], sub_orders[key]
-   end
-end
-
-function util.lua_versions()
-   local versions = { "5.1", "5.2", "5.3" }
+function util.lua_versions(sort)
+   local versions = { "5.1", "5.2", "5.3", "5.4" }
    local i = 0
-   return function()
-      i = i + 1
-      return versions[i]
+   if sort == "descending" then
+      i = #versions + 1
+      return function()
+         i = i - 1
+         return versions[i]
+      end
+   else
+      return function()
+         i = i + 1
+         return versions[i]
+      end
    end
 end
 
+function util.lua_path_variables()
+   local cfg = require("luarocks.core.cfg")
+   local lpath_var = "LUA_PATH"
+   local lcpath_var = "LUA_CPATH"
+
+   local lv = cfg.lua_version:gsub("%.", "_")
+   if lv ~= "5_1" then
+      if os.getenv("LUA_PATH_" .. lv) then
+         lpath_var = "LUA_PATH_" .. lv
+      end
+      if os.getenv("LUA_CPATH_" .. lv) then
+         lcpath_var = "LUA_CPATH_" .. lv
+      end
+   end
+   return lpath_var, lcpath_var
+end
+
 function util.starts_with(s, prefix)
    return s:sub(1,#prefix) == prefix
 end
@@ -431,18 +333,6 @@ function util.printout(...)
    io.stdout:write("\n")
 end
 
---- Print a line to standard error
-function util.printerr(...)
-   io.stderr:write(table.concat({...},"\t"))
-   io.stderr:write("\n")
-end
-
---- Display a warning message.
--- @param msg string: the warning message
-function util.warning(msg)
-   util.printerr("Warning: "..msg)
-end
-
 function util.title(msg, porcelain, underline)
    if porcelain then return end
    util.printout()
@@ -461,11 +351,11 @@ function util.this_program(default)
       cur = dbg.source
       i=i+1
    end
-   return last:sub(2)
+   return last:sub(1,1) == "@" and last:sub(2) or last
 end
 
 function util.deps_mode_help(program)
-   local cfg = require("luarocks.cfg")
+   local cfg = require("luarocks.core.cfg")
    return [[
 --deps-mode=<mode>  How to handle dependencies. Four modes are supported:
                     * all - use all trees from the rocks_trees list
@@ -488,7 +378,7 @@ function util.see_help(command, program)
 end
 
 function util.announce_install(rockspec)
-   local cfg = require("luarocks.cfg")
+   local cfg = require("luarocks.core.cfg")
    local path = require("luarocks.path")
 
    local suffix = ""
@@ -496,8 +386,7 @@ function util.announce_install(rockspec)
       suffix = " (license: "..rockspec.description.license..")"
    end
 
-   local root_dir = path.root_dir(cfg.rocks_dir)
-   util.printout(rockspec.name.." "..rockspec.version.." is now installed in "..root_dir..suffix)
+   util.printout(rockspec.name.." "..rockspec.version.." is now installed in "..path.root_dir(cfg.root_dir)..suffix)
    util.printout()
 end
 
@@ -511,7 +400,7 @@ local function collect_rockspecs(versions, paths, unnamed_paths, subdir)
    local fs = require("luarocks.fs")
    local dir = require("luarocks.dir")
    local path = require("luarocks.path")
-   local deps = require("luarocks.deps")
+   local vers = require("luarocks.core.vers")
 
    if fs.is_dir(subdir) then
       for file in fs.dir(subdir) do
@@ -521,7 +410,7 @@ local function collect_rockspecs(versions, paths, unnamed_paths, subdir)
             local rock, version = path.parse_name(file)
 
             if rock then
-               if not versions[rock] or deps.compare_versions(version, versions[rock]) then
+               if not versions[rock] or vers.compare_versions(version, versions[rock]) then
                   versions[rock] = version
                   paths[rock] = file
                end
@@ -551,6 +440,14 @@ function util.get_default_rockspec()
          return unnamed_paths[1]
       end
    else
+      local fs = require("luarocks.fs")
+      local dir = require("luarocks.dir")
+      local basename = dir.base_name(fs.current_dir())
+
+      if paths[basename] then
+         return paths[basename]
+      end
+
       local rock = next(versions)
 
       if rock then
@@ -566,144 +463,6 @@ function util.get_default_rockspec()
    end
 end
 
--- from http://lua-users.org/wiki/SplitJoin
--- by PhilippeLhoste
-function util.split_string(str, delim, maxNb)
-   -- Eliminate bad cases...
-   if string.find(str, delim) == nil then
-      return { str }
-   end
-   if maxNb == nil or maxNb < 1 then
-      maxNb = 0    -- No limit
-   end
-   local result = {}
-   local pat = "(.-)" .. delim .. "()"
-   local nb = 0
-   local lastPos
-   for part, pos in string.gmatch(str, pat) do
-      nb = nb + 1
-      result[nb] = part
-      lastPos = pos
-      if nb == maxNb then break end
-   end
-   -- Handle the last field
-   if nb ~= maxNb then
-      result[nb + 1] = string.sub(str, lastPos)
-   end
-   return result
-end
-
---- Remove repeated entries from a path-style string.
--- Example: given ("a;b;c;a;b;d", ";"), returns "a;b;c;d".
--- @param list string: A path string (from $PATH or package.path)
--- @param sep string: The separator
-function util.remove_path_dupes(list, sep)
-   assert(type(list) == "string")
-   assert(type(sep) == "string")
-   local parts = util.split_string(list, sep)
-   local final, entries = {}, {}
-   for _, part in ipairs(parts) do
-      part = part:gsub("//", "/")
-      if not entries[part] then
-         table.insert(final, part)
-         entries[part] = true
-      end
-   end
-   return table.concat(final, sep)
-end
-
----
--- Formats tables with cycles recursively to any depth.
--- References to other tables are shown as values.
--- Self references are indicated.
--- The string returned is "Lua code", which can be procesed
--- (in the case in which indent is composed by spaces or "--").
--- Userdata and function keys and values are shown as strings,
--- which logically are exactly not equivalent to the original code.
--- This routine can serve for pretty formating tables with
--- proper indentations, apart from printing them:
--- io.write(table.show(t, "t"))   -- a typical use
--- Written by Julio Manuel Fernandez-Diaz,
--- Heavily based on "Saving tables with cycles", PIL2, p. 113.
--- @param t table: is the table.
--- @param name string: is the name of the table (optional)
--- @param indent string: is a first indentation (optional).
--- @return string: the pretty-printed table
-function util.show_table(t, name, indent)
-   local cart     -- a container
-   local autoref  -- for self references
-
-   local function isemptytable(t) return next(t) == nil end
-   
-   local function basicSerialize (o)
-      local so = tostring(o)
-      if type(o) == "function" then
-         local info = debug.getinfo(o, "S")
-         -- info.name is nil because o is not a calling level
-         if info.what == "C" then
-            return ("%q"):format(so .. ", C function")
-         else 
-            -- the information is defined through lines
-            return ("%q"):format(so .. ", defined in (" .. info.linedefined .. "-" .. info.lastlinedefined .. ")" .. info.source)
-         end
-      elseif type(o) == "number" then
-         return so
-      else
-         return ("%q"):format(so)
-      end
-   end
-   
-   local function addtocart (value, name, indent, saved, field)
-      indent = indent or ""
-      saved = saved or {}
-      field = field or name
-      
-      cart = cart .. indent .. field
-      
-      if type(value) ~= "table" then
-         cart = cart .. " = " .. basicSerialize(value) .. ";\n"
-      else
-         if saved[value] then
-            cart = cart .. " = {}; -- " .. saved[value] .. " (self reference)\n"
-            autoref = autoref ..  name .. " = " .. saved[value] .. ";\n"
-         else
-            saved[value] = name
-            --if tablecount(value) == 0 then
-            if isemptytable(value) then
-               cart = cart .. " = {};\n"
-            else
-               cart = cart .. " = {\n"
-               for k, v in pairs(value) do
-                  k = basicSerialize(k)
-                  local fname = ("%s[%s]"):format(name, k)
-                  field = ("[%s]"):format(k)
-                  -- three spaces between levels
-                  addtocart(v, fname, indent .. "   ", saved, field)
-               end
-               cart = cart .. indent .. "};\n"
-            end
-         end
-      end
-   end
-   
-   name = name or "__unnamed__"
-   if type(t) ~= "table" then
-      return name .. " = " .. basicSerialize(t)
-   end
-   cart, autoref = "", ""
-   addtocart(t, name, indent)
-   return cart .. autoref
-end
-
-function util.array_contains(tbl, value)
-   for _, v in ipairs(tbl) do
-      if v == value then
-         return true
-      end
-   end
-   return false
-end
-
 -- Quote Lua string, analogous to fs.Q.
 -- @param s A string, such as "hello"
 -- @return string: A quoted string, such as '"hello"'
@@ -711,4 +470,51 @@ function util.LQ(s)
    return ("%q"):format(s)
 end
 
+--- Normalize the --namespace flag and the user/rock syntax for namespaces.
+-- If a namespace is given in user/rock syntax, update the --namespace flag;
+-- If a namespace is given in --namespace flag, update the user/rock syntax.
+-- In case of conflicts, the user/rock syntax takes precedence.
+function util.adjust_name_and_namespace(ns_name, flags)
+   assert(type(ns_name) == "string" or not ns_name)
+   assert(type(flags) == "table")
+
+   if not ns_name then
+      return
+   elseif ns_name:match("%.rockspec$") or ns_name:match("%.rock$") then
+      return ns_name
+   end
+
+   local name, namespace = util.split_namespace(ns_name)
+   if namespace then
+      flags["namespace"] = namespace
+   end
+   if flags["namespace"] then
+      name = flags["namespace"] .. "/" .. name
+   end
+   return name:lower()
+end
+
+-- Split name and namespace of a package name.
+-- @param ns_name a name that may be in "namespace/name" format
+-- @return string, string? - name and optionally a namespace
+function util.split_namespace(ns_name)
+   local p1, p2 = ns_name:match("^([^/]+)/([^/]+)$")
+   if p1 then
+      return p2, p1
+   end
+   return ns_name
+end
+
+function util.deep_copy(tbl)
+   local copy = {}
+   for k, v in pairs(tbl) do
+      if type(v) == "table" then
+         copy[k] = util.deep_copy(v)
+      else
+         copy[k] = v
+      end
+   end
+   return copy
+end
+
 return util
